package dragon.rail.engine;

import java.util.LinkedHashMap;

import android.graphics.PointF;
import dragon.rail.engine.shape.Shape;
import dragon.rail.sopwith.SopwithReduxView;
import dragon.rail.sopwith.util.Float2;

public abstract class Renderable {

	protected Float2 position;
	public Control control;
	public Float2 velocity;
	public SopwithReduxView gamePanel;
	public String name;
	public Float2 perceivedVelocity;
	public boolean positionTicked;
	public SafeArrayList<Shape> collision;
	public int collisionFrame;
	public int collisionBits;
	public LinkedHashMap<Renderable, Integer> collidedTicks;
	public Renderable parent;
	public int linkIndexOnParent;
	public int linkFrame;
	public SafeArrayList<Links> links;
	public SafeArrayList<Renderable> children;

	public Renderable(SopwithReduxView gamePanel) {
		control = null;
		position = null;
		velocity = new Float2();
		this.gamePanel = gamePanel;
		collision = new SafeArrayList<Shape>();
		collisionBits = 0;
		collisionFrame = 0;
		collidedTicks = null;
		// numPrevPositions = 3;
		name = "";
		perceivedVelocity = new Float2();
		parent = null;
		linkIndexOnParent = -1;
		linkFrame = 0;
		links = new SafeArrayList<Links>();
		children = new SafeArrayList<Renderable>();
		// prevPositions = initPrevPositions(numPrevPositions);
	}

	public void tick() {
		if (control != null) {
			control.control();
		}
	}

	public void setControl(Control control) {
		this.control = control;
	}

	public Shape getShape() {
		return collision.get(collisionFrame);
	}

	public void collidedWith(Renderable sprite) {
		if (collidedTicks == null) {
			collidedTicks = new LinkedHashMap<Renderable, Integer>();
		}
		Integer integer = collidedTicks.get(sprite);
		if (integer == null) {
			integer = 1;
		} else {
			integer++;
		}
		collidedTicks.put(sprite, integer);
	}

	public boolean getCollisionBit(int bit) {
		return ((collisionBits & (1 << bit)) != 0);
	}

	public void noCollideWith(Renderable sprite) {
		if (collidedTicks == null) {
			return;
		}

		Integer integer = collidedTicks.get(sprite);
		if (integer != null) {
			if (integer == 0) {
				collidedTicks.remove(sprite);
			} else {
				integer = 0;
				collidedTicks.put(sprite, integer);
			}
		}
	}

	public void collide(Renderable sprite, int ticks) {
		if (control != null) {
			if (ticks > 0) {
				control.startCollision(sprite);
				control.collide(sprite, ticks);
			} else {
				control.endCollision(sprite);
			}
		}
	}

	public void setCollisionBit(int bit, boolean state) {
		if (state) {
			this.collisionBits |= 1 << bit;
		} else {
			this.collisionBits &= ~(1 << bit);
		}
	}

	public Float2 getPosition() {
		return position;
	}

	public void setPosition(float x, float y) {
		if (position != null) {
			position.x = x;
			position.y = y;
		} else {
			position = new Float2(x, y);
		}
	}

	public void setPosition(PointF position) {
		this.position.x = position.x;
		this.position.y = position.y;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public void setStaticPosition(float x, float y) {
		position.x = x;
		position.y = y;
		// updatePrevPositions();
	}

	public void setStaticPosition(PointF position) {
		this.position.x = position.x;
		this.position.y = position.y;
		// updatePrevPositions();
	}

	public void setPositionX(float x) {
		this.position.x = x;
	}

	public void setPositionY(float y) {
		this.position.y = y;
	}

	public void move(int xDiff, int yDiff) {
		Float2 position = getPosition();
		setPosition(position.x + xDiff, position.y + yDiff);
	}

	public void perceivedVelocity() {
		// if (numPrevPositions > 0)
		// {
		perceivedVelocity = new Float2(position);
		// perceivedVelocity.subtract(prevPositions[0]);
		// }
	}

	protected boolean isLinked() {
		return parent != null;
	}

	public void linkTo(Renderable parent, int linkIndexOnParent) {
		parent.children.add(this);
		this.parent = parent;
		this.linkIndexOnParent = linkIndexOnParent;

		position = new Float2();
		velocity = new Float2();
	}

	public void unlink() {
		if (isLinked()) {
			velocity = new Float2(parent.velocity);

			parent.children.remove(this);
			parent = null;
			linkIndexOnParent = -1;
		}
	}

	public void updatePosition() {

		if (isLinked()) {
			if (parent.positionTicked) {
				// updatePrevPositions();
				position = parent.getLinkPointInWorldSpace(linkIndexOnParent);
				if (position == null) {
					position = new Float2();
				}
				positionTicked = true;
			}
		} else {

			// updatePrevPositions();
			position.x += velocity.x;
			position.y += velocity.y;
			positionTicked = true;
		}
	}

	public Float2 getLinkPoint(int linkIndex) {
		Links links = getLink();
		if (links == null) {
			return null;
		}
		return links.linkPoints.get(linkIndex);
	}

	protected Float2 getLinkPointInWorldSpace(int linkIndex) {
		Links link = getLink();
		if (link == null) {
			return null;
		}

		Float2 linkPoint = link.getLinkPoint(linkIndex);
		if (linkPoint == null) {
			return null;
		} else {
			return new Float2(linkPoint).add(position);
		}
	}

	public Shape getCollision() {
		return collision.get(collisionFrame);
	}

	public boolean isCollidable() {
		return getCollision() != null;
	}

	private Links getLink() {
		return links.get(linkFrame);
	}

	public void setCollsion(int frame, Shape shape) {
		collision.set(frame, shape);
	}

	public void setLinks(int frame, Links links) {
		this.links.set(frame, links);
	}

	public void setLinks(int frame, int linkIndex, float x, float y) {
		Links links = this.links.get(frame);
		if (links == null) {
			links = new Links(this);
			setLinks(frame, links);
		}
		links.setLinkPoint(linkIndex, x, y);
	}

	public void remove() {
		// gamePanel.removeSprite(this);

	}

	public void prepositionTick() {
		positionTicked = false;
		updatePosition();
	}

	public void postpositionTick() {
		perceivedVelocity();
	}

	public boolean isVisible() {
		if (isLinked()) {
			Float2 float2 = parent.getLinkPoint(linkIndexOnParent);
			if (float2 == null) {
				return false;
			}
		}
		return true;
	}

	public abstract void render(SopwithReduxView gamePanel);

	/**protected Float2[] initPrevPositions(int numPrevPositions) {
		Float2[] positons = new Float2[numPrevPositions];
		if (position == null) {
			for (int i = 0; i < numPrevPositions; i++) {
				positons[i] = null;
			}
		} else {
			for (int i = 0; i < numPrevPositions; i++) {
				positons[i] = new Float2(position);
			}
		}
		return positons;
	}*/

}
