package dragon.rail.sopwith.game;

/*
 * 
 * ArcadeCanvas takes care of the layer between our 
 * game/interface logic and the Android system
 * 
 * It's biggest function is to handle threads and 
 * take care of touchscreen/key events
 * 
 * I use this class mostly because it makes porting between other systems pretty easy
 * By having this class on all other systems to take care of those system-specific stuff
 * we can keep the core game logic seperate
 * 
 */


import java.util.Random;

import android.app.Activity;
import android.content.Context;
import android.content.res.Resources;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.os.Bundle;
import android.util.AttributeSet;
import android.view.GestureDetector;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.view.View;


/**
* The main Arcade Canvas class for Android devices
*
* @author      OrangePixel
* @version     3, 0
* @since       1.0
*/

public class ArcadeCanvas extends SurfaceView implements SurfaceHolder.Callback {

	
	// our main canvas to paint on
	public static Canvas 				theCanvas = null;
	
	// the main paint object, takes care of colors, transparency, flipping,etc
	public static Paint 				myPaint;
	

	// Various default GAMESTATE value's
	public final static int 			INSPLASH = 0;											
	public final static int 			INMENU = 1;
	public final static int 			INGAME = 2;
	public final static int 			INLOADER = 3;
	
	// The current gamestate (menu,ingame,inhighscores,etc)
	public int 							GameState;

	// We need random values! and this makes it easier
	public static Random 				randomGenerator;
	
	// Surface manager, used for drawing etc
	private SurfaceHolder 				mSurfaceHolder;

	// used to signal a paused state
	public boolean 						paused;

	// see if the back-key is pressed
	public boolean 						backPressed;
	public boolean 						backLocked;


	// a global world-tickers.. runs from 0..100
	public static int 					worldTicks;
	

	// set to true if the touchscreen was touched
	public static boolean 				touched;
	
	// this is set to true if touch-release event happens
	// code has to set it to false when a touch happens and you want to "lock" it
	public static boolean 				touchReleased;

	// location of the last touch
	public static float 				touchX;
	public static float 				touchY;

	// used for touch-scroll or touch-flick events
	public static boolean 				touchScrolled;
	public static boolean 				touchFlinged;
	public static float 				touchDistanceX;
	public static float 				touchDistanceY;
	
	// a link to our parent Activity
	private Activity 					myActivity;
	
	// used for limiting our game loop-speed
	private long 						loopStart;
	private long 						loopEnd;
	public long 						loopWait;

	// Alpha value used for splash fadein/fadeout 
	private static int 					splashAlpha;


	// record a copy of the width and height.. slightly faster to use this then calling methods
	public static int					displayW;
	public static int					displayH;

	
	
	/**
	 * Fetch the Alpha value of the splash logo (incremented automatically)
	 */
	public final static int getSplashAlpha() {
		return splashAlpha;
	}
	
	
	
	/**
	 * Set WorldTicks to a specific value
	 */
	public final static void setWorldTicks(int newtick) {
		worldTicks=newtick;
	}
	
	
	
	/**
	 * Get current WorldTicks value
	 */
	public final static int getWorldTicks() {
		return worldTicks;
	}
	
	
	
	
	
	
	/**
	 * Reset the touch values, code calls this after a touch is processed
	 */
	public final static void resetTouch() {
		touchX=-1;
		touchY=-1;
		touched=false;
		touchReleased=true;
		
		touchScrolled=false;
		touchFlinged=false;
		touchDistanceX=0;
		touchDistanceY=0;
	}
	
	
	
	
	
	
	/**
	 * Retrieve a random number between DIGIT1 and DIGIT2
	 *
	 * @param digit1 The minimum value
	 * @param digit2 The maximum value
	 */
	public final static int getRandom(int digit1) {
		return  randomGenerator.nextInt(digit1);
	}

	  
	  

	
		/**
		 * Main game thread 
		 * taking care of calling logicloop, gameloop, etc
		 * @author Pascal
		 *
		 */
		
		class MainThread extends Thread {
			// is it running or not!?!!
			public boolean 				isRunning = false;
		
			// main constructor
		    public MainThread(SurfaceHolder surfaceHolder) {
				mSurfaceHolder=surfaceHolder;
		    }
		    
		
		    /**
		     * Run method, runs until isRunning is set to false
		     */
		    public void run() {
		    	// the amount of time to sleep
				int 				fpsSleep;
				
				while (isRunning) {
		
			    	
			    	loopStart = System.currentTimeMillis();
					
			    	worldTicks++;																// global worldticking
					if (worldTicks>1000) worldTicks=0;											// heart-beat of the game
		
					theCanvas=null;
		            try {
		            	theCanvas = mSurfaceHolder.lockCanvas(null);
		            	
		            	displayW=theCanvas.getWidth();
		            	displayH=theCanvas.getHeight();
		                
		                synchronized (mSurfaceHolder) {
		
		                	
		            		switch (GameState) {
		            			case INGAME:
		            				GameLoop();
		            			break;
		            			
		            			case INSPLASH:
		            				SplashLoop();
		            				splashAlpha+=16; 
		            				if (splashAlpha>255) init();
		            			break;
		            			
		            			default:
		            				LogicLoop();
		            			break;
		            		}
		
		            		
		                }
		            } finally {
		                // do this in a finally so that if an exception is thrown
		                // during the above, we don't leave the Surface in an
		                // inconsistent state
		                if (theCanvas != null) {
		                    mSurfaceHolder.unlockCanvasAndPost(theCanvas);
		                }
		            }
		
					loopEnd = System.currentTimeMillis();
					
					
		
					// limit gamespeed to a nice frame rate 
					// (best is 40-60fps)  
					// math example: (1000/32fps = 31ms per frame)
					if (loopEnd-loopStart<50)	
						fpsSleep=50-(int)(loopEnd-loopStart); 
					else 
						fpsSleep=2;  // incase we are trailing behind fps
		
					try {
						sleep(fpsSleep);
					} catch (InterruptedException e) {}
		
		        }
		    }        
		
		    
		    
		}
// End MainThread class







		/**
		 * Touch thread, handles the interface stuff (keys,touching,etc)
		 * @author Pascal
		 */
		class TouchThread extends Thread {
			public boolean 			isRunning = false;
		
		    public TouchThread() {
		    }
		
		    
		
		    @Override
		    public void run() {
				
				while (isRunning) {
					try {
						sleep(64);
					} catch (InterruptedException e) {}
		
		        }
		    }        
		
		    
		    
		    /**
		     * Handles a key event.
		     * 
		     * @param keyCode the key that was pressed
		     * @param msg the original event object
		     * @return true if handled, false if not handled
		     */
		    boolean onKeyEvent(int keyCode, KeyEvent msg) {
		    	if (msg.getAction()==KeyEvent.ACTION_DOWN) {
			    	if (keyCode==KeyEvent.KEYCODE_BACK) {
			    		if (!backLocked) backPressed=true;
			    		return true;
			    	}
		    	}
		    	
		        if (msg.getAction()==KeyEvent.ACTION_UP) {
		        	if (keyCode==KeyEvent.KEYCODE_MENU){ 
		        		paused=true;
		        		return true;
		        	}

		
		        	if (keyCode==KeyEvent.KEYCODE_BACK) {
		        		backPressed=false;
		        		backLocked=false;
		        	}
		        }
		        return false;
		    }
		
		    
		    
		    
		    boolean onTouchEvent(MotionEvent event) {
		    	if (event.getAction()==MotionEvent.ACTION_DOWN || event.getAction()==MotionEvent.ACTION_MOVE) {
			    	touchX=event.getX();
			    	touchY=event.getY();
			    	touched=true;
		    	} 
		    	
		    	if (event.getAction()==MotionEvent.ACTION_UP) {
		    		touched=false;
		    		touchReleased=true;
		    		touchX=-1;
		    		touchY=-1;
		    	}
		    	
		    	gestureDetector.onTouchEvent(event); 
			    	
	    		try {
	    			sleep(100);
	    		} catch (InterruptedException e) { 
	    		}
		    	
		    	return true;
		    }
		    
		    
		}








	// we need a link to our resources, this one is always available
	public Resources 				res = this.getResources();

	// The thread that actually draws the animation
	public MainThread 				gameThread;
	public TouchThread 				touchThread;
	public GestureDetector 			gestureDetector;


	
	/**
	 * Fetches the animation thread 
	 * 
	 * @return the animation thread
	 */
	public MainThread getThread() {
	    return gameThread;
	}




	/**
	 * Main constructor for our Arcade Canvas
	 * @param context 	- android specific, context to run in 
	 * @param attrs		- android specific, attributes, we don't really use it
	 */
	public ArcadeCanvas(Context context, AttributeSet attrs) {
		super(context, attrs);
	
	    // register our interest in hearing about changes to our surface
	    SurfaceHolder holder = getHolder();
	    	holder.addCallback(this);

	    // initialise some important values
	    backLocked=false;
	    backPressed=false;
        
	    
	    // initialise our Paint object
	    // this is the "filter" we use on our draw-actions
	    // contains colors, sizes, flipping, alpha, etc,etc.
	    myPaint=new Paint();

		// start Alpha value of our splash image
		splashAlpha=1;
		
		
		// start in our Splash screen state
		GameState=INSPLASH;

		// init our game loop 
		loopWait=System.currentTimeMillis();										// used by splash 
		
		// initialise our random-number generator
		randomGenerator = new Random();											// global random generator

		
		// setup our touchlisteners, to pass it on to our touch thread
	    this.setOnTouchListener(new OnTouchListener() {
	        public boolean onTouch(View v, MotionEvent event) {
	            return touchThread.onTouchEvent(event);
	        }
	    });

	    // ditto for the keylistener
	    this.setOnKeyListener(new OnKeyListener() {
	    	public boolean onKey(View v, int keyCode, KeyEvent keyEvent) {
	    		return touchThread.onKeyEvent(keyCode,keyEvent);
	    	}
	    });
	    
	    
	    // easy Android way of finding scroll or fling moves on the touchscreen
	    this.gestureDetector= new GestureDetector(
	    		new GestureDetector.SimpleOnGestureListener() {
	    			
	    			// Scrolling of the background
	    			 public final boolean onScroll(MotionEvent e1, MotionEvent e2, float distanceX, float distanceY) {
	    				touchScrolled=true;
	    				touchDistanceX=-distanceX;
	    				touchDistanceY=-distanceY;
	    				
	    				return true;
	    			}
	    			
	    			// Fling / swipe movements on the screen
	    			public final boolean onFling(final MotionEvent e1,
	    				final MotionEvent e2, final float velocityX, final float velocityY) {
	    				touchFlinged=true;
	    				touchDistanceX=velocityX;
	    				touchDistanceY=velocityY;

	    				return true;
	    			}

	    		}
	    ); 
	    
	    // initialise our Touch variables
	    resetTouch();

	    // make sure we receive key events
		setFocusable(true);
		// and touch events
		setFocusableInTouchMode(true); 
	}


	
	// application is paused by system, save important settings 
	//  - overwritten in the maincanvas
    public Bundle saveState(Bundle saveState) {
        return saveState;
    }
    
	// application is continued again, so reload our settings
    // - overwrite in the maincanvas
    public void restoreState(Bundle savedState) {
    }






    /* Callback invoked when the surface dimensions change. */
    public void surfaceChanged(SurfaceHolder holder, int format, int width,
        int height) {
    }


    /* Callback invoked when the Surface has been created and is ready to be used. */
    public void surfaceCreated(SurfaceHolder holder) {
    	// start the thread here so that we don't busy-wait in run()
    	// waiting for the surface to be created
    	// create thread only; it's started in surfaceCreated()
    	gameThread = new MainThread(holder);
		touchThread= new TouchThread();    	

    	// initialise our loop and running values
		loopWait=System.currentTimeMillis();
		gameThread.isRunning=true;
		touchThread.isRunning=true;
		
		// and start the threads!
		if (!gameThread.isAlive())	gameThread.start();
		if (!touchThread.isAlive()) touchThread.start();
    }


	/*
	 * Callback invoked when the Surface has been destroyed and must no longer
	 * be touched. WARNING: after this method returns, the Surface/Canvas must
	 * never be touched again!
	 */
	public void surfaceDestroyed(SurfaceHolder holder) {
	    // we have to tell thread to shut down & wait for it to finish, or else
	    // it might touch the Surface after we return and explode
	    boolean retry = true;
	    touchThread.isRunning=false;
	    gameThread.isRunning=false;
	    while (retry) {
	        try {
	        	touchThread.join();
	            retry = false;
	        } catch (InterruptedException e) {
	        }
	    }
	
	    // same with main gamethread
	    while (retry) {
	        try {
	        	gameThread.join();
	            retry = false;
	        } catch (InterruptedException e) {
	        }
	    }
	    
	    touchThread=null;
	    gameThread=null;
	}





	/**
	 * Keep a link to our parent Activity
	 * 
	 * @param parentActivity
	 */
	public void setParentActivity(Activity parentActivity) {
		myActivity=parentActivity;
	}
	
	public Activity getParentActivity() {
		return myActivity;
	}
	

	// various game logic stuff, overwrite in maincanvas
    public void init() {}
	public void LogicLoop() { }
	public void SplashLoop() { }
	public void GameLoop(){ }

	


}
